/**
 * @file c2n.c
 * The main program
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2001,2002,2004,2006 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/** Program version */
#define VERSION "1.2.1"
/** Release date (day.month.year) */
#define RELEASE_DATE "03.07.2007"

#ifdef __AMIGA__
/** version information */
static const char version[] = "$VER: C2N " VERSION " (" RELEASE_DATE ")";
#endif /* __AMIGA */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "c2n.h"
#ifdef USE_SERIAL
# include "serial.h"
#endif /* USE_SERIAL */
#ifdef USE_PCM
# ifdef USE_TAP
#  include "pcm_in.h"
# endif /* USE_TAP */
# include "pcm_out.h"
#endif /* USE_PCM */
#include "decode.h"
#include "encode.h"
#include "oric_d.h"
#include "oric_e.h"
#include "kim_d.h"
#include "kim_e.h"

/** Flag: verbose mode */
unsigned verbose = 0;

/** Number of bytes read from stdin */
static long stdinpos;
/** Name of the current file */
static const char* currentFile;
/** Maximum number of encountered errors */
static unsigned error_threshold = 10;
/** Number of encountered errors */
static unsigned error_count;

/** Error reporter
 * @param pec	the unexpected pulse or error code, or character (| 256)
 * @param block	the block in which the error occurred (0=first header)
 * @param pos	number of bytes decoded from the stream
 * @return	nonzero on fatal error
 */
static unsigned
error_report (unsigned pec, unsigned block, unsigned pos)
{
  if (currentFile)
    fputs (currentFile, stderr), fputs (": ", stderr);
  switch (pec) {
  case Pause:
    fputs ("unexpected break in the pulse stream", stderr);
    break;
  case Short:
    fputs ("unexpected short pulse", stderr);
    break;
  case Medium:
    fputs ("unexpected medium pulse", stderr);
    break;
  case Long:
    fputs ("unexpected long pulse", stderr);
    break;
  case OutOfSync:
    fputs ("out of synchronisation", stderr);
    break;
  case Parity:
    fputs ("parity error", stderr);
    break;
  case Checksum:
    fputs ("block checksum mismatch", stderr);
    break;
  case NoFirst:
    fputs ("first copy is missing", stderr);
    break;
  case NoSecond:
    fputs ("second copy is missing", stderr);
    break;
  case LongBlock:
    fprintf (stderr, "block %u is %u bytes longer than expected\n",
	     block, pos);
    goto threshold;
  case ShortBlock:
    fprintf (stderr, "block %u is %u bytes shorter than expected\n",
	     block, pos);
    goto threshold;
  case Unexpected:
    fprintf (stderr, "unexpected header tag 0x%02x in block %u\n",
	     pos, block);
    goto threshold;
  case Mismatch:
    fputs ("mismatch between first and second copy", stderr);
    break;
  case Countdown:
    fprintf (stderr, "unrecognized countdown byte 0x%02x at block %u\n",
	     pos, block);
    goto threshold;
  default:
    if (pec > 255)
      fprintf (stderr, "unexpected character 0x%02x", pec & 0xff);
    else
      fputs ("unexpected internal error", stderr);
    break;
  }
  fprintf (stderr, " at %u decoded bytes of block %u\n", pos, block);

 threshold:
  if (++error_count > error_threshold) {
    fprintf (stderr, "more than %u errors; aborted\n", error_threshold);
    return 1;
  }
  return 0;
}

/** Cleanup function */
static void
cleanup (void)
{
#ifdef HAS_SERIAL
  serial_cleanup ();
#endif /* HAS_SERIAL */
#ifdef USE_PCM
  pcm_cleanup ();
# ifdef USE_TAP
  pcm_in_cleanup ();
# endif /* USE_TAP */
#endif /* USE_PCM */
}

/** Signal handler
 * @param num	the signal number
 */
static void
sig (int num)
{
  cleanup ();
  signal (num, SIG_DFL);
  raise (num);
}

/** Read a pulse from standard input
 * @return	the read pulse
 */
static enum pulse
p_read (void)
{
  int c = getchar ();
  stdinpos++;
  switch (c) {
  case EOF:
    stdinpos--;
    /* fall through */
  case 'A':
    return Pause;
  case 'B':
    return Short;
  case 'C':
    return Medium;
  case 'D':
    return Long;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n", c);
  return Pause;
}

#ifdef USE_TAP
/** short pulse width in a tape image (unit=8 s) */
static int tap_short;
/** medium pulse width in a tape image (unit=8 s) */
static int tap_medium;
/** long pulse width in a tape image (unit=8 s) */
static int tap_long;

/** Read a pulse from a tape image
 * @return	the read pulse
 */
static enum pulse
tap_read (void)
{
  int c = getchar ();
  stdinpos++;
  if (c == EOF)
    stdinpos--;
  else if (c > 0 && c <= tap_short)
    return Short;
  else if (c <= tap_medium)
    return Medium;
  else if (c <= tap_long)
    return Long;
  return Pause;
}
# ifdef USE_PCM
/** Read a pulse from an audio file
 * @return	the read pulse
 */
static enum pulse
wav_read (void)
{
  /** pulse width read last (EOF=none) */
  static int c0 = EOF;
  int c = c0;
  c0 = EOF;
  if (c != EOF)
    goto already;
  for (;;) {
    c = pcm_in_read ();
  already:
    if (c == EOF && !pcm_in_next ())
      continue;
    if (c == EOF)
      return Pause;
    if (c <= tap_short)
      return Short;
    if (c <= tap_medium)
      return Medium;
    if (c <= tap_long)
      return Long;
    if (verbose)
      fprintf (stderr, "-i: skipping pause at offset %ld\n", pcm_in_pos ());
    /* got a long pulse: lump following long pulses as a single pause */
    do
      c0 = pcm_in_read ();
    while (c0 != EOF && c0 > tap_long);
    if (verbose)
      fprintf (stderr, "-i: skipped pause until offset %ld\n", pcm_in_pos ());

    return Pause;
  }
}
# endif /* USE_PCM */
#endif /* USE_TAP */

#ifndef HAS_SERIAL
/** Write a pulse to the serial line or to standard output
 * @param p	the pulse to be written
 */
static void
p_write (enum pulse p)
{
  putchar (p);
}
#endif /* !HAS_SERIAL */

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful
 */
int
main (int argc, char** argv)
{
  /** mode of operation */
  enum { Encode,
#ifdef USE_TAP
	 DecodeTAP,
# ifdef USE_PCM
	 DecodeWAV,
# endif /* USE_PCM */
#endif /* USE_TAP */
	 Decode
  } mode = Encode;
  /** tape format type */
  enum { CBM = 0, KIM1, PLUS4, ORIC1 } format = CBM;
  /** flag: raw operation (input and output pulse streams) */
  int raw = 0;
#ifdef HAS_SERIAL
  /** flag: fast operation (unsafe with USB RS-232 adapters) */
  int fast = 0;
  /** C2N232 hardware version */
  int c2n232_version = 0;
#endif /* HAS_SERIAL */
  /** program name */
  const char* prog = *argv;
  /** short pulse width (unit=8 microseconds) */
  static unsigned pulse_short = 0;
  /** medium pulse width (unit=8 microseconds) */
  static unsigned pulse_medium = 0;
  /** long pulse width (unit=8 microseconds) */
  static unsigned pulse_long = 0;
  /** initial sync pulse count */
  static unsigned sync_begin = 0;
  /** intra-block sync pulse count */
  static unsigned sync_intra = 0;

  atexit (cleanup);
  signal (SIGINT, sig);

  /* process the options */
  for (argv++; argc > 1 && **argv == '-'; argv++, argc--) {
    const char* opts = *argv;

    if (!strcmp (opts, "--")) { /* disable processing further options */
      argv++;
      argc--;
      break;
    }

    while (*++opts) /* process all flags */
      switch (*opts) {
#ifdef HAS_SERIAL
      case 'f':
	fast = 1;
	break;
      case 'c':
	if (argc <= 2 || serial_open)
	  goto Usage;
	if (serial_init (*++argv, fast, &c2n232_version))
	  return 2;
	argc--;
	break;
#endif /* HAS_SERIAL */
      case 'd':
	mode = Decode;
	break;
#ifdef USE_TAP
      case 'D':
	mode = DecodeTAP;
	break;
#endif /* USE_TAP */
      case 'e':
	mode = Encode;
	break;
#ifdef USE_PCM
      case 'E':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  /** sample rate: negative=inverted signal */
	  int rate = strtol (*++argv, &r, 0);
	  /** number of channels (1=mono, 2=stereo) */
	  unsigned channels = 1;
	  /** flag: nonzero=output RIFF WAVE (.wav) header */
	  int riff = 0;
	  /** flag: nonzero=signed output, 0=unsigned */
	  int sign = 1;
	  /** sample file */
	  FILE* f = 0;
	  /** sample format */
	  enum pcm_type format = rate > 0 ? eight : eight_;
	  samplerate = rate > 0 ? rate : -rate;
	  pcm_cleanup ();
	  argc--;
	  if (!rate) {
	  EUsage:
	    fprintf (stderr, "-E: [+-]rate[blsuwW]*[,file] expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	Enext:
	  switch (*r++) {
	  case 0:
	    break;
	  case ',':
	    if (!*r)
	      goto EUsage;
	    if (!(f = fopen (r, "wb"))) {
	      fputs ("-E ", stderr);
	      fputs (*argv, stderr);
	      perror (": fopen");
	      return 2;
	    }
	    break;
	  case 'b':
	    format = rate > 0 ? msb16 : msb16_;
	    riff = 0;
	    goto Enext;
	  case 'W':
	    riff = 1;
	    sign = 0;
	    /* fall through */
	  case 'l':
	    format = rate > 0 ? lsb16 : lsb16_;
	    goto Enext;
	  case 'w':
	    format = rate > 0 ? eight : eight_;
	    riff = sign = 1;
	    goto Enext;
	  case 's':
	    channels = 2;
	    goto Enext;
	  case 'u':
	    sign = 0;
	    goto Enext;
	  default:
	    goto EUsage;
	  }
	  samplerate = rate > 0 ? rate : -rate;
	  pcm_init (f, format, channels, riff, sign);
	  mode = Encode;
	}
	break;
# ifdef USE_TAP
      case 'i':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  /** threshold: negative=inverted signal */
	  int threshold = strtol (*++argv, &r, 0);
	  /** DC offset */
	  int bias = 128;
	  /** sample file */
	  FILE* f = 0;
	  pcm_in_cleanup ();
	  argc--;
	  if (!*argv || r == *argv || threshold < -127 || threshold > 127) {
	  iUsage:
	    fprintf (stderr, "-i: [+-]threshold[+offset][,file.wav] expected,"
		     " got %s\n", *argv);
	    goto Usage;
	  }
	inext:
	  switch (*r++) {
	  case 0:
	    break;
	  case ',':
	    if (!*r)
	      goto iUsage;
	    if (!(f = fopen (r, "rb"))) {
	      fputs ("-i ", stderr);
	      fputs (*argv, stderr);
	      perror (": fopen");
	      return 2;
	    }
	    break;
	  case '+':
	    bias = strtol (r, &r, 0);
	    if (bias >= 1 && bias <= 254)
	      goto inext;
	  default:
	    goto iUsage;
	  }
	  pcm_in_init (f, bias, threshold);
	  mode = DecodeWAV;
	}
	break;
# endif /* USE_TAP */
#endif /* USE_PCM */
      case 'h':
	goto Usage;
      case '1':
	format = ORIC1;
	break;
      case '2':
	format = PLUS4;
	break;
      case 'k':
	format = KIM1;
	break;
      case 'p':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  pulse_short = strtoul (s, &r, 0);
	  if (*r != ',') {
	  pUsage:
	    fprintf (stderr, "-p: short,medium,long expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	  s = r + 1;
	  pulse_medium = strtoul (s, &r, 0);
	  if (*r != ',')
	    goto pUsage;
	  s = r + 1;
	  pulse_long = strtoul (s, &r, 0);
	  if (*r)
	    goto Usage;
	}
	break;
      case 'r':
	raw = 1;
	break;
      case 's':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* r;
	  char* s = *++argv;
	  argc--;
	  sync_begin = strtoul (s, &r, 0);
	  if (!*r);
	  else if (*r != ',') {
	  sUsage:
	    fprintf (stderr, "-s: begin[,intra] expected, got %s\n",
		     *argv);
	    goto Usage;
	  }
	  else {
	    s = r + 1;
	    sync_intra = strtoul (s, &r, 0);
	    if (*r)
	      goto sUsage;
	  }
	}
	break;
      case 't':
	if (argc <= 2)
	  goto Usage;
	else {
	  char* s;
	  error_threshold = strtoul (*++argv, &s, 0);
	  argc--;
	  if (*s) {
	    fprintf (stderr, "-t: non-numeric argument %s\n", *argv);
	    goto Usage;
	  }
	}
	break;
      case 'v':
	verbose++;
	break;

      default:
	goto Usage;
      }
  }

  if (argc < 2) {
  Usage:
    fprintf (stderr,
	     "C2N " VERSION
	     " - Commodore C2N tape format encoder and decoder\n"
	     "Usage: %s [options] file(s)\n", prog);
    fputs ("Options: -1: use Oric-1 pulse widths and format\n"
	   "         -2: use Commodore 264 series pulse widths and format\n"
	   "         -k: use KIM-1 Hypertape format\n"
#ifdef HAS_SERIAL
	   "         -c device: specify the name of the C2N232 device\n"
	   "         -f: enable fast mode (unsafe with USB RS-232 adapters)\n"
#endif /* HAS_SERIAL */
	   "         -d: decode pulses to binary stream\n"
#ifdef USE_TAP
	   "         -D: decode tape image to binary stream\n"
#endif /* USE_TAP */
	   "         -e: encode binary stream to pulses (default)\n"
#ifdef USE_PCM
	   "         -E [+-]rate[blsuwW]*[,file]: produce PCM audio\n"
# ifdef USE_TAP
	   "         -i [+-]threshold[+offset][,file.wav]: decode audio\n"
# endif /* USE_TAP */
#endif /* USE_PCM */
	   "         -h: request this help text\n"
	   "         -p short,medium,long: specify pulse lengths for C2N232\n"
	   "         -r: enable raw operation (input and output pulses)\n"
	   "         -s begin,intra: specify lengths of sync pulse streams\n"
	   "         -t number: specify maximum number of errors\n"
	   "         -v: enable verbose operation\n"
	   "         --: Stop processing any further options.\n",
	   stderr);
    return 1;
  }

#ifdef USE_PCM
  /* Compute the sample sequences for the pulses */
  if (samplerate) {
    do {
      switch (format) {
      case CBM:
	if (!pcm_compute_samples (pulse_short ? pulse_short : 45,
				  pulse_medium ? pulse_medium : 63,
				  pulse_long ? pulse_long : 83))
	  continue;
	break;
      case KIM1:
	if (!pcm_compute_samples (pulse_short ? pulse_short : 34,
				  pulse_medium ? pulse_medium : 52,
				  0))
	  continue;
	break;
      case PLUS4:
	if (!pcm_compute_samples (pulse_short ? pulse_short : 60,
				  pulse_medium ? pulse_medium : 120,
				  pulse_long ? pulse_long : 240))
	  continue;
	break;
      case ORIC1:
	if (!pcm_compute_samples (pulse_short ? pulse_short : 52,
				  pulse_medium ? pulse_medium : 104,
				  0))
	  continue;
	break;
      }
      fputs ("-E: failed to compute PCM samples\n", stderr);
      return 2;
    }
    while (0);
  }
#endif /* USE_PCM */

#ifdef HAS_SERIAL
  /* Set the pulse widths */
  if (serial_open) {
    unsigned char cmd[5], *op = cmd;
    switch (mode) {
    case Encode:
      *op++ = 3;
      *op++ = 255; /* pause duration */
      switch (format) {
      case CBM:
	*op++ = pulse_short ? pulse_short >> 1 : 22;
	*op++ = pulse_medium ? pulse_medium >> 1 : 31;
	*op++ = pulse_long ? pulse_long >> 1 : 41;
	break;
      case KIM1:
	*op++ = pulse_short ? pulse_short >> 1 : 17;
	*op++ = pulse_medium ? pulse_medium >> 1 : 26;
	*op++ = 255; /* no long pulses */
	break;
      case PLUS4:
	*op++ = pulse_short ? pulse_short >> 1 : 30;
	*op++ = pulse_medium ? pulse_medium >> 1 : 60;
	*op++ = pulse_long ? pulse_long >> 1 : 120;
	break;
      case ORIC1:
	*op++ = pulse_short ? pulse_short >> 1 : 26;
	*op++ = pulse_medium ? pulse_medium >> 1 : 39;
	*op++ = 255; /* no long pulses */
	break;
      }
      break;
# ifdef USE_TAP
    case DecodeTAP:
      fputs ("interpreting -D as -d in the presence of -c\n", stderr);
      mode = Decode;
#  ifdef USE_PCM
      goto decode;
    case DecodeWAV:
      fputs ("interpreting -i as -d in the presence of -c\n", stderr);
      mode = Decode;
    decode:
#  endif /* USE_PCM */
# endif /* USE_TAP */
    case Decode:
      *op++ = 2;
      switch (format) {
      case CBM:
	*op++ = pulse_short ? pulse_short : 54;
	*op++ = pulse_medium ? pulse_medium : 74;
	*op++ = pulse_long ? pulse_long : 94;
	break;
      case KIM1:
	*op++ = pulse_short ? pulse_short : 40;
	*op++ = pulse_medium ? pulse_medium : 48;
	*op++ = pulse_long ? pulse_long : 60;
	break;
      case PLUS4:
	*op++ = pulse_short ? pulse_short : 100;
	*op++ = pulse_medium ? pulse_medium : 180;
	*op++ = pulse_long ? pulse_long : 250;
	break;
      case ORIC1:
	*op++ = pulse_short ? pulse_short >> 1 : 54;
	*op++ = pulse_medium ? pulse_medium >> 1 : 80;
	*op++ = 255; /* no long pulses */
	break;
      }
      break;
    }

    if (c2n232_version) {
      /* Scale the pulse widths for the lower sample rate. */
      unsigned char* c;
      for (c = op - 3; c < op && *c != 255; c++)
	*c *= .576; /* 18.432 MHz/256 instead of 8 MHz/64 */
    }

    if (serial_command (cmd, op - cmd))
      return 2;
    if (mode == Encode &&
	serial_timeout ())
      return 2;
  }
#endif /* HAS_SERIAL */

#if defined USE_TAP && defined USE_PCM
  if (mode == DecodeWAV && pcm_in_next ()) {
    fputs ("-i: RIFF WAVE PCM file expected\n", stderr);
    return 2;
  }
#endif /* USE_TAP && USE_PCM */

  /* Process the files. */

  for (; --argc; argv++) {
    /** input file */
    FILE* df;
    /** pulse input stream */
    pulse_r_t rd;
    switch (mode) {
    case Encode:
      if (!(df = fopen (currentFile = *argv, "rb"))) {
	fputs (currentFile, stderr);
	perror (": fopen (read)");
	continue;
      }

#ifdef HAS_SERIAL
      serial_empty ();
#endif /* HAS_SERIAL */

      if (raw) {
	int c;
	while ((c = fgetc (df)) != EOF) {
	  enum pulse p = c;
	  switch (c) {
	  case Pause: case Short: case Medium:
	    break;
	  case Long:
	    if (format == CBM || format == PLUS4 || format == KIM1)
	      break;
	  default:
	    fprintf (stderr, "%s: unrecognized pulse character 0x%02x\n",
		     currentFile, c);
	    continue;
	  }
#ifdef USE_PCM
	  if (samplerate)
	    p_write_pcm (p);
	  else
#endif /* USE_PCM */
	    p_write (p);
	}
      }
      else if (format == ORIC1)
	encode_oric (df, error_report,
#ifdef USE_PCM
		     samplerate ? p_write_pcm :
#endif /* USE_PCM */
		     p_write,
		     sync_begin ? sync_begin : 50,
		     sync_intra ? sync_intra : 100);
      else if (format == KIM1)
	encode_kim (df, error_report,
#ifdef USE_PCM
		    samplerate ? p_write_pcm :
#endif /* USE_PCM */
		    p_write,
		    sync_begin ? sync_begin : 20,
		    sync_intra ? sync_intra : 1);
      else if (format == PLUS4)
	encode264 (df, error_report,
#ifdef USE_PCM
		   samplerate ? p_write_pcm :
#endif /* USE_PCM */
		   p_write,
		   sync_begin ? sync_begin : 1500,
		   sync_intra ? sync_intra : 16384);
      else
	encode (df, error_report,
#ifdef USE_PCM
		samplerate ? p_write_pcm :
#endif /* USE_PCM */
		p_write,
		sync_begin ? sync_begin : 1500,
		sync_intra ? sync_intra : 5376);
      fclose (df);
#ifdef HAS_SERIAL
      if (serial_open)
	serial_flush ();
#endif /* HAS_SERIAL */
      break;
#ifdef USE_TAP
# ifdef USE_PCM
    case DecodeWAV:
# endif /* USE_PCM */
    case DecodeTAP:
      switch (format) {
      case ORIC1:
	tap_short = pulse_short ? pulse_short >> 1 : 54;
	tap_medium = pulse_medium ? pulse_medium >> 1 : 80;
	tap_long = tap_medium;
	break;
      case CBM:
	tap_short = pulse_short ? pulse_short : 54;
	tap_medium = pulse_medium ? pulse_medium : 74;
	tap_long = pulse_long ? pulse_long : 94;
	break;
      case KIM1:
	tap_short = pulse_short ? pulse_short : 40;
	tap_medium = pulse_medium ? pulse_medium : 48;
	tap_long = pulse_long ? pulse_long : 60;
	break;
      case PLUS4:
	tap_short = pulse_short ? pulse_short : 100;
	tap_medium = pulse_medium ? pulse_medium : 180;
	tap_long = pulse_long ? pulse_long : 250;
	break;
      }
      /* fall through */
#endif /* USE_TAP */
    case Decode:
      if (!(df = fopen (currentFile = *argv, "wb"))) {
	fputs (currentFile, stderr);
	perror (": fopen (write)");
	continue;
      }

      error_count = 0;
#ifdef HAS_SERIAL
      serial_empty ();
#endif /* HAS_SERIAL */
      stdinpos = 0L;

      rd =
#ifdef HAS_SERIAL
	serial_open ? sp_read :
#endif /* HAS_SERIAL */
	(
#if defined USE_TAP && defined USE_PCM
	 mode == DecodeWAV ? wav_read :
#endif /* USE_TAP && USE_PCM */
	 (
#ifdef USE_TAP
	  mode == DecodeTAP ? tap_read :
#endif /* USE_TAP */
	  p_read));

      if (raw) {
	/** maximum number of successive pauses that will be tolerated */
	static const int maxpauses = 5;
	/** number of pauses that will be tolerated */
	int pauses = maxpauses;

	for (;;) {
	  enum pulse p = (*rd) ();
	  if (p == Pause) {
	    if (pauses--)
	      continue;
	    break;
	  }
	  else if (p >= Pause && p <= LongLong) {
	    if (pauses != maxpauses)
	      putc (Pause, df);
	    pauses = maxpauses;
	    putc (p, df);
	  }
	}
      }
      else if (format == KIM1)
	fprintf (stderr, "%s: decoded %u bytes (KIM-1)\n", currentFile,
		 decode_kim (rd, error_report, df));
      else if (format == ORIC1)
	fprintf (stderr, "%s: decoded %u bytes (ORIC-1)\n", currentFile,
		 decode_oric (rd, error_report, df));
      else if (format == PLUS4)
	fprintf (stderr, "%s: decoded %u bytes (plus/4)\n", currentFile,
		 decode264 (rd, error_report, df));
      else
	fprintf (stderr, "%s: decoded %u bytes\n", currentFile,
		 decode (rd, error_report, df));
      fclose (df);
      if (verbose
#ifdef HAS_SERIAL
	  && rd != sp_read
#endif /* HAS_SERIAL */
	  )
	fprintf (stderr, "read %ld bytes from input\n",
		 rd == wav_read ? pcm_in_pos () : stdinpos);
      break;
    }
  }

#ifdef HAS_SERIAL
  if (serial_open)
    serial_go_idle ();
#endif /* HAS_SERIAL */

  return 0;
}
