/**
 * @file kim_d.c
 * KIM-1 tape decoding functions
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2004 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include <stdio.h>
#include <stdlib.h>

#include "c2n.h"
#include "kim_d.h"

/* In the Hypertape format,
 * an 'I' symbol is represented by 3 short pulses, and
 * an 'O' symbol is represented by 2 long pulses.
 * The original KIM-1 format uses 9 and 6 pulses, respectively.
 * Byte boundaries are not marked in any way - the decoder will
 * expect to find a long enough sequence of SYN (^V, 026) bytes.
 * There are no parity bits either.  There is a 16-bit checksum
 * for the whole file.  The data bytes are represented with hexadecimal
 * numbers.
 */

/* The KIM-1 format is insensitive to the polarity of the signal, because
 * the polarity is reversed at each 'I' symbol.  Due to bandwidth limitations
 * of the C2N232 serial interface, the C2N232 firmware reports only the time
 * intervals between high-to-low transitions of the signal.  Thus, we see three
 * different symbols for the two, 's' (short pulse) and 'l' (long pulse):
 *
 * Short = ss
 * Long = ll
 * s Medium = ssl
 * l Medium = lls
 *
 * In order to properly decode medium-length cycles reported by the C2N232,
 * the decoder must remember whether the last pulse decoded was short or long.
 */

/** the pulse stream reader */
static pulse_r_t dec_rd;

/** Decode the next pulse
 * @return	Short, Long or Pause
 */
static enum pulse
decode_pulse (void)
{
  /** flag: was a short half-wave observed last? */
  static int lastShort;
  /** flag: has a half-wave been buffered? */
  static int buffered;
  /** a pulse read from the tape */
  enum pulse p;

  if (buffered) {
    buffered = 0;
    return lastShort ? Short : Long;
  }

  p = (*dec_rd) ();
  switch (p) {
  case Short: /* ss */
    lastShort = 1;
    break;
  case Medium: /* ls or sl */
    p = lastShort ? Short : Long;
    lastShort = !lastShort;
    break;
  case Long: /* ll */
    lastShort = 0;
    break;
  case Pause:
    lastShort = 0;
    return p;
  }
  buffered = 1;
  return p;
}

/** Decode a data symbol
 * @return	0 for '0', 1 for '1', or <0 on error
 */
static int
decode_bit (void)
{
  /** flag: got start bit already? */
  static int start;
  /** pulse counter: negative='O', positive='I' */
  int cnt;
  /** number of subsequent pauses */
  unsigned pauses = 0;

  /* 0 = IIO, 1 = IOO */
  /* I = Short+, O = Long+ */

  if (start) {
    start = 0;
    goto waitStop;
  }

  /* wait for 'I' symbol (start bit) */
  for (;;) {
    switch (decode_pulse ()) {
    default:
      if (++pauses == 5)
	return -1 - Pause;
      break;
    case Long:
      pauses = 0;
      break;
    case Short:
      pauses = 0;
      goto waitStop;
    }
  }

 waitStop:
  /* wait for 'O' symbol (data or stop bit) */
  for (cnt = 2;;) {
    switch (decode_pulse ()) {
    default:
      if (++pauses == 5)
	return -1 - Pause;
      break;
    case Long:
      pauses = 0;
      goto waitNext;
    case Short:
      pauses = 0;
      cnt += 2;
      break;
    }
  }

 waitNext:
  /* wait for 'I' symbol (start bit of next byte) or timeout */
  while ((cnt -= 3) > 0) {
    switch (decode_pulse ()) {
    default:
      if (++pauses == 5)
	return -1 - Pause;
      break;
    case Short:
      /* set flag: got the start bit of next byte */
      start = 1;
      /* got longer period of 'I' than 'O' => IIO = '0' */
      return 0;
    case Long:
      pauses = 0;
      break;
    }
  }
  /* timeout: got longer period of 'O' than 'I' => IOO = '1' */
  return 1;
}

/** Decode a data character
 * @return	the character, or <0 on error
 */
static int
decode_char (void)
{
  int c, cnt;
  for (cnt = 8, c = 0; cnt--; ) {
    int bit = decode_bit ();
    c >>= 1;
    if (bit < 0)
      return bit;
    else if (bit)
      c |= 0x80;
  }
  return c & 255;
}

/** Decode a SYN character followed by 10 SYN characters
 * @return	zero on success, <0 on error
 */
static int
decode_sync (void)
{
  int b, c;
 resync:
  for (c = 0; c != 0x16; ) {
    b = decode_bit ();
    if (b < 0)
      return b;
    c >>= 1;
    if (b < 0)
      return b;
    if (b)
      c |= 0x80;
  }
  for (c = 10; c--; ) {
    b = decode_char ();
    if (b < 0)
      return b;
    if (b != 0x16)
      goto resync;
  }
  return 0;
}

/** Decode a hexadecimal digit
 * @param out	the data output stream
 * @return	the digit (0..15), or 256 | char, or <0 on error
 */
static int
decode_hex (FILE* out)
{
  int c = decode_char ();
  if (c < 0) /* error */
    return c;
  putc (c, out);
  if (c >= 0x30 && c <= 0x39) /* '0'..'9' */
    return c - 0x30;
  if (c >= 0x41 && c <= 0x46) /* 'A'..'F' */
    return c - (0x41 - 10);
  return c | 256;
}

/** Decode a hexadecimally represented byte
 * @param out	the data output stream
 * @return	the byte (0..255), or 256 | char, or <0 on error
 */
static int
decode_byte (FILE* out)
{
  int h, l;
  h = decode_hex (out);
  if (h & ~15)
    return h;
  l = decode_hex (out);
  if (l & ~15)
    h = 0;
  return h << 4 | l;
}

/** KIM-1 pulse stream decoder
 * @param rd	the pulse stream reader
 * @param err	the error reporter
 * @param out	the data output stream
 * @return	number of bytes converted
 */
unsigned
decode_kim (pulse_r_t rd, pulse_error_t err, FILE* out)
{
  /** number of the block being decoded */
  unsigned block = 0;
  /** total data payload byte count */
  unsigned cnt = 0;
  /** byte count in block (SYN or payload bytes) */
  unsigned bcnt;
  /** data character */
  int c;
  /** computed checksum */
  unsigned cksum;

  dec_rd = rd;
  cnt = 0;

 sync:
  if (verbose)
    fputs ("skipping a sync mark", stderr), fflush (stderr);

 resync:
  /* synchronise on the SYN bytes */
  if (decode_sync ()) {
    if (verbose)
      fputs ("... none found\n", stderr);
    return cnt;
  }
  bcnt = 11;

  /* got 11 SYN bytes: accept SYN bytes, wait for '*' */
  for (;;) {
    c = decode_char ();
    bcnt++;
    switch (c) {
    case 0x16:	/* SYN */
      break;
    case 0x2a:	/* '*' */
      bcnt--; /* do not include the '*' in statistics */
      goto data;
    default:	/* other: resync */
      goto resync;
    }
  }

  /* got '*': read the data */
 data:
  if (verbose)
    fprintf (stderr, " (%u SYN bytes)\n", bcnt);

  /* omit the SYN bytes, output the '*' only */
  putc (0x2a, out);

  for (cksum = bcnt = 0;; cnt++, bcnt++) {
    c = decode_byte (out);
    if (c == (256 | 0x2f)) /* end of data? */
      break;
    else if (c < 0 || (c & 256)) {
      if (!err || (*err) (c < 0 ? -1 - c : c, block, cnt))
	return cnt;
    }
    else {
      /* compute the checksum from all correctly received bytes
       * but the first one in a block (file ID)
       */
      if (bcnt)
	cksum += c;
    }
  }

  /* got '/': read the checksum */
  c = decode_byte (out);
  if (c < 0 || (c & 256)) {
    if (!err || (*err) (c < 0 ? -1 - c : c, block, cnt))
      return cnt;
  }
  if ((unsigned) c != (cksum & 0xff)) {
    if (!err || (*err) (Checksum, block, cnt))
      return cnt;
  }
  c = decode_byte (out);
  if (c < 0 || (c & 256)) {
    if (!err || (*err) (c < 0 ? -1 - c : c, block, cnt))
      return cnt;
  }
  if ((unsigned) c != (cksum >> 8 & 0xff)) {
    if (!err || (*err) (Checksum, block, cnt))
      return cnt;
  }
  block++;
  goto sync;
}
