/**
 * @file kim_e.c
 * KIM-1 tape encoding functions
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2004 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include <stdio.h>
#include <stdlib.h>

#include "c2n.h"
#include "kim_e.h"

/* In the Hypertape 6 format,
 * an 'I' symbol is represented by 3 short half-waves, and
 * an 'O' symbol is represented by 2 long half-waves.
 * The original KIM-1 format uses 18 and 12 half-waves, respectively.
 * Byte boundaries are not marked in any way - the decoder will
 * expect to find a long enough sequence of SYN (^V, 026) bytes.
 * There are no parity bits either.  There is a 16-bit checksum
 * for the whole file.  The data bytes are represented with hexadecimal
 * numbers.
 */

/** The pulse writer */
static pulse_w_t enc_wr;
/** The transfer speed (1=fastest Hypertape, 6=KIM-1) */
static unsigned enc_rate;
/** Status: last pulse written */
static enum { NONE, SHORT, LONG } lastPulse;

/** Encode a data symbol
 * @param sym	0 for 'O', nonzero for 'I'
 */
static void
encode_sym (unsigned sym)
{
  /** number of half-waves to write */
  unsigned cnt;
  if (sym) {
    cnt = 3 * enc_rate;
    switch (lastPulse) {
    case NONE:
      break;
    case SHORT:
      cnt++;
      break;
    case LONG:
      (*enc_wr) (MediumShort);
      cnt--;
      break;
    }
    /* write 2 half-waves at a time */
    for (; cnt > 2; cnt -= 2)
      (*enc_wr) (Short);
    lastPulse = cnt ? SHORT : NONE;
  }
  else {
    cnt = 2 * enc_rate;
    switch (lastPulse) {
    case NONE:
      break;
    case LONG:
      cnt++;
      break;
    case SHORT:
      (*enc_wr) (ShortMedium);
      cnt--;
      break;
    }
    /* write 2 half-waves at a time */
    for (; cnt > 2; cnt -= 2)
      (*enc_wr) (Medium);
    lastPulse = cnt ? LONG : NONE;
  }
}

/**
 * Encode a bit
 * @param bit	the bit to be encoded
 */
static void
encode_bit (unsigned bit)
{
  encode_sym (1);
  encode_sym (!bit);
  encode_sym (0);
}

/** Encode a data character
 * @param c	the character
 */
static void
encode_char (int c)
{
  int cnt;
  for (cnt = 8; cnt--; c >>= 1)
    encode_bit (c & 1);
}

/** Read and encode a hexadecimal digit
 * @param in	the data stream
 * @return	the digit, or c | 256 for unexpected characters, or EOF
 */
int
read_hex (FILE* in)
{
  int c = getc (in);
  if (c < 0) /* error */
    return c;
  encode_char (c);
  if (c >= 0x30 && c <= 0x39) /* '0'..'9' */
    return c - 0x30;
  if (c >= 0x41 && c <= 0x46) /* 'A'..'F' */
    return c - (0x41 - 10);
  return c | 256;
}

/** Read and encode a hexadecimally represented byte
 * @param in	the data stream
 * @return	the byte (0..255), or 256 | char, or EOF
 */
int
read_byte (FILE* in)
{
  int h, l;
  h = read_hex (in);
  if (h & ~15)
    return h;
  l = read_hex (in);
  if (l & ~15)
    h = 0;
  return h << 4 | l;
}

/** KIM-1 pulse stream encoder
 * @param in	the data stream
 * @param err	the error reporter
 * @param wr	the pulse stream writer
 * @param begin	number of initial synchronization bytes
 * @param rate	baud rate: 1=Hypertape, 6=original KIM-1
 * @return	number of blocks written; 0 on failure
 */
unsigned
encode_kim (FILE* in, pulse_error_t err, pulse_w_t wr,
	    unsigned begin, unsigned rate)
{
  /** current block */
  unsigned block;
  enc_wr = wr;
  enc_rate = rate;

  for (block = 0;; block++) {
    int c;
    /** data count */
    unsigned cnt;
    /** block checksum */
    unsigned cksum;

    /* read and skip SYN and EOF bytes, expect '*' */
    for (;;) {
      c = fgetc (in);
      switch (c) {
      case EOF:
	return block; /* end of input: return */
      default:
	if (!err || !(*err) (Unexpected, block, c))
	  return block;
	/* fall through */
      case 0x04:
      case 0x16:
	continue; /* ignore EOF and SYN */
      case 0x2a:
	break; /* got '*' */
      }
      /* got '*': exit the loop */
      break;
    }

    /* encode the SYN bytes */
    if (verbose)
      fprintf (stderr, "encoding a sync mark (%u bytes)\n", begin);
    for (cnt = begin; cnt--; )
      encode_char (0x16);

    /* encode the '*' mark */
    encode_char (0x2a);

    for (cnt = cksum = 0;; cnt++) {
      c = read_byte (in);
      if (c == (256 | 0x2f)) /* end of data? */
	break;
      else if (c == EOF) {
      eof:
	if (err)
	  (*err) (ShortBlock, block, cnt);
	return block;
      }
      else if (c & 256) {
	if (!err || !(*err) (c, block, cnt))
	  return block;
      }
      else {
	/* compute the checksum from all correctly received bytes
	 * but the first one in a block (file ID)
	 */
	if (cnt)
	  cksum += c;
      }
    }

    if (verbose)
      fprintf (stderr, "encoded %u bytes\n", cnt);

    /* got '/': read the checksum */
    c = read_byte (in);
    if (c == EOF)
      goto eof;
    if (c & 256) {
      if (!err || (*err) (c, block, cnt))
	return block;
    }
    if ((unsigned) c != (cksum & 0xff)) {
      if (!err || (*err) (Checksum, block, cnt))
	return block;
    }
    c = read_byte (in);
    if (c == EOF)
      goto eof;
    if (c & 256) {
      if (!err || (*err) (c, block, cnt))
	return block;
    }
    if ((unsigned) c != (cksum >> 8 & 0xff)) {
      if (!err || (*err) (Checksum, block, cnt))
	return block;
    }
    encode_char (0x04); /* EOF */
    encode_char (0x04); /* EOF */
    (*enc_wr) (Pause);
  }
  return block;
}
