/**
 * @file pcm_out.c
 * Pulse Code Modulated audio output
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2002,2004 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#ifdef USE_PCM
# include "c2n.h"
# include "pcm_out.h"
# include <stdlib.h>
# include <math.h>

/** A PCM sample of a pulse */
struct pcm_pulse
{
  char* data;		/**< the data bytes */
  unsigned long half;	/**< offset of half wave position (180 degrees) */
  unsigned long length;	/**< length of the data bytes */
};

/** sample rate in Hz */
unsigned samplerate;
/** bytes per sample per channel */
static unsigned samplewidth;
/** zero level of the samples */
static short samplebias;
/** number of output channels */
static unsigned pcm_channels;
/** number of bytes written to a WAV file (-1L == no WAV header) */
static long wavbytes = -1L;

/** PCM samples of the pulses */
static struct pcm_pulse pcm[3];
/** The selected PCM sample format */
static enum pcm_type pcm_format;
/** File descriptor for PCM output (NULL=use stdout) */
static FILE* pcm_file;

/** Compute a stream of pulse code modulation (PCM) samples for a pulse
 * @param sample	the sample stream to be computed
 * @param width		the cycle time/8s
 * @return		the computed sample, NULL on failure
 */
static struct pcm_pulse*
pcm_compute (struct pcm_pulse* sample,
	     unsigned width)
{
  double arg = 0, inc;
  char* c;
  short mult = 0;

  sample->length = width;
  sample->length *= samplerate;
  sample->length /= 125000;
  sample->half = sample->length / 2;
  inc = 2 * 3.14159265358979323846 / sample->length;
  sample->half *= pcm_channels;
  sample->length *= pcm_channels;

  switch (pcm_format) {
  case eight:
    mult = 127; break;
  case eight_:
    mult = -127; break;
  case lsb16: case msb16:
    sample->half *= 2; sample->length *= 2; mult = 32767; break;
  case lsb16_: case msb16_:
    sample->half *= 2; sample->length *= 2; mult = -32767; break;
  }

  free (sample->data);
  if (!(sample->data = malloc (sample->length)))
    return 0;

  for (c = sample->data; c < sample->data + sample->length; arg += inc) {
    short s = samplebias + mult * sin (arg);
    unsigned ch = pcm_channels;
    switch (pcm_format) {
    case eight: case eight_:
      while (ch--) *c++ = s;
      break;
    case lsb16: case lsb16_:
      while (ch--) { *c++ = s; *c++ = s >> 8; }
      break;
    case msb16: case msb16_:
      while (ch--) { *c++ = s >> 8; *c++ = s; }
      break;
    }
  }
  return sample;
}

/** Write an initial WAV header to the PCM output file */
static void
write_wav_header (void)
{
  /** start of WAVE header for PCM data */
  static const unsigned char header1[22] = {
    0x52, 0x49, 0x46, 0x46, /* "RIFF" */
    0xff, 0xff, 0xff, 0xff, /* length of "RIFF" block (initially unknown) */
    0x57, 0x41, 0x56, 0x45, /* "WAVE" */
    0x66, 0x6d, 0x74, 0x20, /* "fmt " */
    0x10, 0x00, 0x00, 0x00, /* length of "fmt " block: 16 bytes */
    0x01, 0x00 /* 1 (PCM) */
  };
  /** end of WAVE header for PCM data */
  static const unsigned char header2[8] = {
    0x64, 0x61, 0x74, 0x61, /* "data" */
    0xff, 0xff, 0xff, 0xff, /* length of remaining file (initially unknown) */
  };

  /** output file */
  FILE* f = pcm_file ? pcm_file : stdout;
  /** bytes per second */
  unsigned long datarate = (unsigned long)
    samplerate * samplewidth * pcm_channels;

  fwrite (header1, sizeof header1, 1, f);
  /* number of channels */
  putc (pcm_channels, f);
  putc (pcm_channels >> 8, f);
  /* samples per second */
  putc (samplerate, f);
  putc (samplerate >> 8, f);
  putc (samplerate >> 16, f);
  putc (samplerate >> 24, f);
  /* bytes per second */
  putc (datarate, f);
  putc (datarate >> 8, f);
  putc (datarate >> 16, f);
  putc (datarate >> 24, f);
  /* bytes per sample */
  putc (samplewidth * pcm_channels, f);
  putc ((samplewidth * pcm_channels) >> 8, f);
  /* bits per channel */
  putc (samplewidth << 3, f);
  putc (samplewidth >> 5, f);
  fwrite (header2, sizeof header2, 1, f);
}

/** Adjust the length of a WAV header written to the PCM output file */
static void
adjust_wav_header (void)
{
  FILE* f = pcm_file ? pcm_file : stdout;
  long bytes = wavbytes + 40L;
  wavbytes = 0L;
  if (fseek (f, -bytes, SEEK_CUR))
    return;
  bytes -= 4L;
  putc (bytes, f);
  putc (bytes >> 8, f);
  putc (bytes >> 16, f);
  putc (bytes >> 24, f);
  if (fseek (f, 32L, SEEK_CUR))
    return;
  bytes -= 36L;
  putc (bytes, f);
  putc (bytes >> 8, f);
  putc (bytes >> 16, f);
  putc (bytes >> 24, f);
  fseek (f, bytes, SEEK_CUR);
}

/** Set up the PCM resources
 * @param out		output file handle (NULL=use stdout)
 * @param format	selected PCM sample format
 * @param channels	number of channels (1=mono, 2=stereo)
 * @param riff		flag: nonzero=output in RIFF WAVE format
 * @param sign		flag: nonzero=signed output, 0=unsigned
 * @param 
 */
void
pcm_init (FILE* out,
	  enum pcm_type format,
	  unsigned channels,
	  int riff,
	  int sign)
{
  samplebias = 0;

  switch (format) {
  case eight: case eight_:
    samplewidth = 1;
    if (riff || !sign) /* 8-bit RIFF WAVE is unsigned */
      samplebias = -128;
    break;
  case msb16: case msb16_:
    if (riff)
      format--; /* RIFF WAVE is LSB first */
    /* fall through */
  case lsb16: case lsb16_:
    samplewidth = 2;
    if (!riff && sign) /* 16-bit RIFF WAVE is signed */
      samplebias = -32768;
    break;
  }

  pcm_file = out;
  pcm_format = format;
  pcm_channels = channels;
  wavbytes = riff ? 0L : -1L;

  if (riff)
    write_wav_header ();
}

/** Compute the PCM samples
 * @param t_short	short cycle time/8s
 * @param t_medium	medium cycle time/8s
 * @param t_long	long cycle time/8s
 * @return		zero on success, nonzero on failure
 */
int
pcm_compute_samples (unsigned t_short,
		     unsigned t_medium,
		     unsigned t_long)
{
  if (samplerate &&
      (!t_short || pcm_compute (&pcm[0], t_short)) &&
      (!t_medium || pcm_compute (&pcm[1], t_medium)) &&
      (!t_long || pcm_compute (&pcm[2], t_long)))
    return 0;
  pcm_cleanup ();
  return 1;
}

/** Deallocate the PCM resources */
void
pcm_cleanup (void)
{
  if (wavbytes != -1L) {
    adjust_wav_header ();
    wavbytes = -1L;
  }
  if (pcm_file) {
    fclose (pcm_file);
    pcm_file = 0;
  }
  free (pcm[0].data), free (pcm[1].data), free (pcm[2].data);
  pcm[0].data = pcm[1].data = pcm[2].data = 0;
  samplerate = 0;
}

/** Write audio samples corresponding to a pulse to standard output
 * @param p	the pulse to be written
 */
void
p_write_pcm (enum pulse p)
{
  const struct pcm_pulse* p_pos;
  const struct pcm_pulse* p_neg;
  FILE* f = pcm_file ? pcm_file : stdout;
  int wave;

  wave = (p - Pause) & 3;
  p_pos = wave ? &pcm[wave - 1] : 0;
  wave = (p - Pause) >> 2 & 3;
  if (!wave) {
    /* symmetric wave */
    if (p_pos) {
      fwrite (p_pos->data, p_pos->length, 1, f);
      if (wavbytes != -1L)
	wavbytes += p_pos->length;
    }
    return;
  }
  p_neg = wave ? &pcm[wave - 1] : 0;

  if (p_pos) {
    fwrite (p_pos->data, p_pos->half, 1, f);
    if (wavbytes != -1L)
      wavbytes += p_pos->half;
  }
  if (p_neg) {
    fwrite (p_neg->data + p_neg->half, p_neg->length - p_neg->half, 1, f);
    if (wavbytes != -1L)
      wavbytes += p_neg->length - p_neg->half;
  }
}
#endif /* USE_PCM */
