/**
 * @file serialp.c
 * The POSIX serial line handler
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2001,2002,2004,2006 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include "c2n.h"
#include "serial.h"
#ifdef HAS_SERIAL
/** The name of the serial device */
static const char* serialdev;
# define _BSD_SOURCE /* usleep(), CRTSCTS */
# include <stdio.h>
# include <string.h>
# include <errno.h>
# include <termios.h>
# include <unistd.h>
# include <fcntl.h>
/** file descriptor for the serial device (-1=none) */
int serialfd = -1;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;
/** Report an error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  fprintf (stderr, "c2n -c %s: ", serialdev);
  perror (op);
}

/** input/output buffer */
static char buf[8192];
/** number of bytes read from or written to buf */
static unsigned datain;
/** number of bytes in buf */
static unsigned datalen;
/** nonzero=fast mode (unsafe with USB RS-232 adapters) */
static int serial_fast;

/** Set up the serial communications
 * @param dev		serial device driver name
 * @param fast		nonzero=fast mode (unsafe with USB RS-232 adapters)
 * @param version	(out) hardware version (0 or 1)
 * @return		zero on success, nonzero on failure
 */
int
serial_init (const char* dev, int fast, int* version)
{
  serial_fast = fast;

  if ((serialfd =
       open (serialdev = dev, O_RDWR | O_NONBLOCK | O_NOCTTY)) < 0) {
    report_error ("open");
    return 1;
  }
  else {
    int flags = fcntl (serialfd, F_GETFL);
    if (flags < 0) {
      report_error ("fcntl(GETFL)");
      goto c2n232fail;
    }
    else if (flags & O_NONBLOCK &&
	     fcntl (serialfd, F_SETFL, flags & ~O_NONBLOCK) < 0) {
      report_error ("fcntl(SETFL)");
      goto c2n232fail;
    }
  }

  if (tcgetattr (serialfd, &tio)) {
    report_error ("tcgetattr");
  c2n232fail:
    close (serialfd);
    serialfd = -1;
    return 2;
  }

  memcpy (&newtio, &tio, sizeof newtio);
  newtio.c_iflag = IGNBRK | IGNPAR | IXON;
  newtio.c_oflag = 0;
# ifdef CRTSCTS
  if (!fast)
    newtio.c_cflag = CS8 | CLOCAL | CREAD | CRTSCTS;
  else
# endif
    newtio.c_cflag = CS8 | CLOCAL | CREAD;
  newtio.c_lflag = 0;
  memset (newtio.c_cc, 0, sizeof newtio.c_cc);
  newtio.c_cc[VSTART] = 021; /* DC1, ctrl-q */
  newtio.c_cc[VSTOP] = 023; /* DC3, ctrl-s */
  newtio.c_cc[VMIN] = 0; /* non-blocking read */
  newtio.c_cc[VTIME] = 1; /* 1/10th second timeout */
  cfsetispeed (&newtio, B38400), cfsetospeed (&newtio, B38400);
  tcflush (serialfd, TCIOFLUSH);
  tcsendbreak (serialfd, 0);
  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    report_error ("tcsetattr");
    goto c2n232fail;
  }
  else {
    char last = 0;
    /* send a version identification request and await a response */
    write (serialfd, "\11", 1);
    for (;;) {
      int len = read (serialfd, buf, sizeof buf);
      if (!len) {
	if (last >= 0x30) {
	  *version = last >= 0x60;
	  newtio.c_cc[VMIN] = 0; /* read at least 0 characters */
	  newtio.c_cc[VTIME] = 20; /* wait 2 seconds for end of stream */
	  if (!tcsetattr (serialfd, TCSANOW, &newtio))
	    break;
	  report_error ("tcsetattr");
	}
	else if (last || !errno) {
	  fputs ("c2n -c ", stderr);
	  fputs (serialdev, stderr);
	  fputs (last
		 ? ": unexpected response from C2N232 device\n"
		 : ": no response from C2N232 device\n", stderr);
	}
	else
	  report_error ("read");
	goto c2n232fail;
      }
      last = buf[len - 1];
    }
  }
  return 0;
}

void
serial_cleanup (void)
{
  if (serialfd != -1) {
    tcsetattr (serialfd, TCSADRAIN, &tio);
    tcsendbreak (serialfd, 0);
    close (serialfd);
    serialfd = -1;
  }
}

/** Read a pulse from the serial line
 * @return	the read pulse
 */
enum pulse
sp_read (void)
{
  char c;
  if (!serial_open)
    return Pause;
 again:
  if (datain < datalen)
    c = buf[datain++];
  else {
    ssize_t len = read (serialfd, buf, 1);
    if (len <= 0) {
      serial_cleanup ();
      return Pause;
    }
    datain = 0;
    datalen = (unsigned) len;
    goto again;
  }
  switch (c) {
  case Pause: case Short: case Medium: case Long:
    return c;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n",
	   (unsigned char) c);
  goto again;
}

/** Empty the serial buffers */
void
serial_empty (void)
{
  datain = 0, datalen = 0;
}

/** Flush the serial output buffer */
void
serial_flush (void)
{
}

/** Send a command to the C2N232 device
 * @param buf	the command block
 * @param len	the length of the block
 * @return	zero on success, nonzero on failure
 */
int
serial_command (unsigned char* buf,
		unsigned len)
{
  if ((ssize_t) len != write (serialfd, buf, len)) {
    report_error ("write");
    return 1;
  }
  if (!read (serialfd, buf + 1, 1)) {
    if (errno)
      report_error ("read");
    else
      fputs ("no response from C2N232 device\n", stderr);
    return 2;
  }

  if ((buf[1] & ~0x30) == buf[0])
    return 0;
  fprintf (stderr, "C2N232 responded 0x%02x; expected 0x%02x\n",
	   buf[1], buf[0] | 0x30);
  return 3;
}

/** Disable the serial read timeout
 * @return	zero on success, nonzero on failure
 */
int
serial_timeout (void)
{
  newtio.c_cc[VTIME] = 0; /* no waiting at reading */
  if (tcsetattr (serialfd, TCSANOW, &newtio)) {
    report_error ("tcsetattr");
    return 1;
  }
  return 0;
}

/** Make the C2N232 go idle */
void
serial_go_idle (void)
{
  if (tcdrain (serialfd))
    report_error ("tcdrain");
  if (!serial_fast)
    usleep (250000);
  write (serialfd, "", 1);
}

/** Write a pulse to the serial line or to standard output
 * @param p	the pulse to be written
 */
void
p_write (enum pulse p)
{
  if (serial_open) {
    buf[datain++] = p;
    if (p == Pause || datain == sizeof buf) {
      write (serialfd, buf, datain);
      datain = 0;
    }
  }
  else
    putchar (p);
}
#endif /* HAS_SERIAL */
